using System;
using System.Collections;
using System.ComponentModel;
using iAnywhere.Data.UltraLite;

namespace iAnywhere.Data.UltraLite.DataGridDataSource
{

    /// <summary><para>This class manages the various columns and rows in the table/result-set.
    /// It only retrieves table rows on an as needed basis. The table implements 2
    /// interfaces. Implmenting the IList interface gives this class the ability to
    /// retrieve rows from the database on an as needed basis, and to inform of the
    /// DataSource of the actual number of rows - even though they are not stored in memory.
    /// The ITypedList interface allows this class to convey information about the
    /// column schema to the DataSource. Please also look at the iBindingList interface for
    /// further functionality.
    /// <para></summary>
    internal class CustomTable : IList, ITypedList
    {
        private PropertyDescriptorCollection _propertyDescriptors; // the collection of property descriptors
        private ULDataReader _table; // Ultralite .NET Table or ResultSet
        private ULCursorSchema _schema; // schema that is associated with the casted Cursor
        private string _tableName; // table/result-set name
        private int _currentRowPosition; // currently selected row within table
        private int _numberOfRows; // rows in table - cached b/c the row count operation is expensive

        // constructor
        internal CustomTable(ULDataReader table)
        {
            _table = table;

            // This is the only part of the code that differentiates between a ULTable, a ULResultSet, and a ULDataReader
            // This generality is only possible since this data source is read-only, to match the read-only nature
            // of the DataGrid in the Compact Framework. This
            // is reasonable since a read-only table is essentially a result-set.
            
            // the cursor is a Table
            if (_table is iAnywhere.Data.UltraLite.ULTable)
            {
                _schema = ((iAnywhere.Data.UltraLite.ULTable)_table).Schema;
                _tableName = _schema.Name;
            }
            else if (_table is iAnywhere.Data.UltraLite.ULResultSet)
            // the cursor is a ResultSet
            {
                _schema = ((iAnywhere.Data.UltraLite.ULResultSet)_table).Schema;
                _tableName = "ULResultSet";
            }
            else if (_table is iAnywhere.Data.UltraLite.ULDataReader)
            // the cursor is a DataReader
            {
                _schema = ((iAnywhere.Data.UltraLite.ULDataReader)_table).Schema;
                _tableName = "ULDataReader";
            }
            else
            {
                throw new ULException(ULSQLCode.SQLE_DATATYPE_NOT_ALLOWED, "DataGridDataSource", null, null);
            }

            InitFromTable();
        }

        ~CustomTable()
        {
        }

        // initializes the table wrapper
        private void InitFromTable()
        {

            _table.MoveFirst();
            _currentRowPosition = 0;
            _numberOfRows = (int)_table.RowCount; // caches the number of rows in table

            short numberOfColumns = _schema.ColumnCount;
            String colName;
            Type colType;

            PropertyDescriptor[] pdArray = new PropertyDescriptor[numberOfColumns];

            // creates a list of column descriptors (that emulate properties)
            for (short i = 0; i < numberOfColumns; i++)
            {
                short colID = (short)(i);
                colName = _schema.GetColumnName(colID);
                colType = ULDNETBridgeHelper.GetNETType(_schema.GetColumnULDbType(colID));
                if (colType != null)
                {
                    pdArray[i] = new CustomTableColumn(i, colName, colType);
                }
                else
                {
                    throw new ULException(ULSQLCode.SQLE_DATATYPE_NOT_ALLOWED, "DataGridDataSource", null, null);
                }
            }

            _propertyDescriptors = new PropertyDescriptorCollection(pdArray);
        }

        // gets a new blank row that conforms to the specified column schema
        private CustomTableRow NewRow()
        {

            return new CustomTableRow(_schema.ColumnCount);
        }

        // ITypedList method
        // retrieves the descriptors that allows the DataGrid to discover the schema of the table
        // and query about the data values. Each of the Property Descriptors is associated with
        // a particular column within the table. The DataGrid can query each descriptor about the value
        // contained within the column, given a particular row.
        PropertyDescriptorCollection ITypedList.GetItemProperties(PropertyDescriptor[] listAccessors)
        {

            return _propertyDescriptors;
        }

        // ITypedList method
        // gets the name of the UltraLite Table/ResultSet that this class encapsulates. This name is used to
        // identify the table when external manipulations are done using DataGrid methods.
        // ie. table styles.
        string ITypedList.GetListName(PropertyDescriptor[] listAccessors)
        {

            return _tableName;
        }

        // IList method
        // indicates that the size of the table is fixed and cannot be changed
        public bool IsFixedSize
        {
            get
            {
                return true;
            }
        }

        // IList method
        // indicates that the table rows are read-only
        public bool IsReadOnly
        {
            get
            {
                return true;
            }
        }

        // IList method
        // allows the DataGrid to retrieve a particular row given the row number
        // the set method is not implemented since this table is read-only
        public object this[int index]
        {
            get
            {
                if (index >= _numberOfRows || index < 0)
                {
                    throw new ArgumentOutOfRangeException();
                }

                // moves to the correct point within the table
                _table.MoveRelative(index - _currentRowPosition);
                _currentRowPosition = index;

                CustomTableRow row = NewRow();

                // populates the blank row with the correct values
                if (!ULDNETBridgeHelper.GetCurrentRowData(_table, _schema, row))
                {
                    throw new ULException(ULSQLCode.SQLE_DATATYPE_NOT_ALLOWED, "DataGridDataSource", null, null);
                }

                return row;
            }
            set
            {
                throw new NotSupportedException();
            }
        }

        // IList method
        // Tells the DataGrid about the actual size of the table. Even though, the rows
        // are retrieved on an as needed basis, the proper number of rows needs to be
        // known by the DataGrid. ie. to adjust the scroll bar
        public int Count
        {
            get
            {
                return _numberOfRows;
            }
        }

        // IList -> ICollection method
        // indicates that this class is not thread safe
        public bool IsSynchronized
        {
            get
            {
                return false;
            }
        }

        // Unsupported IList methods

        public int Add(object value)
        {
            throw new NotSupportedException();
        }

        public void Clear()
        {
            throw new NotSupportedException();
        }

        public bool Contains(object value)
        {
            throw new NotSupportedException();
        }

        public int IndexOf(object value)
        {
            throw new NotSupportedException();
        }

        public void Insert(int index, object value)
        {
            throw new NotSupportedException();
        }

        public void Remove(object value)
        {
            throw new NotSupportedException();
        }

        public void RemoveAt(int index)
        {
            throw new NotSupportedException();
        }

        public void CopyTo(Array array, int index)
        {
            throw new NotSupportedException();
        }

        public object SyncRoot
        {
            get
            {
                throw new NotSupportedException();
            }
        }

        public IEnumerator GetEnumerator()
        {
            throw new NotSupportedException();
        }
    }
}   
